from sqlalchemy import *

from dataProviderBase import *
from T0Mon.utils import *
from T0Mon.analysis import *
from T0Mon.consts import *
from T0Mon.sqlManager import *

from T0Mon.globals import logger
import T0Mon.config

### Module for classes providing data about T0 runs ###

class RunConfigKeyProvider(TableDataProvider):
    col1 = [Col("Dataset"), Cols([])]
    col2 = [Col("Repack"), Cols([Col("Proc Version"), Col("HLTDebug")])]
    col3 = [Col("Reco"), Cols([Col("Reco"), Col("CMSsw Version"), Col("Global tag"), Col("Config URL", decorator=buildWebLink), Col("Proc Version")])]
    columns = []
    columns.append(col1)
    columns.append(col2)
    columns.append(col3)

    cols = DoubleCols(columns)

    def __init__(self, runid):
        self.runid = runid
        self.title = "Run " + str(self.runid)
        TableDataProvider.__init__(self)
        TableDataProvider._load(self)

    def _buildQuery(self):
        repackConfig = Tbl(TABLES.REPACK_CONFIG, True)
        recoConfig = Tbl(TABLES.RECO_CONFIG, True)
        primaryDataset = Tbl(TABLES.PRIMARY_DATASET, True)
        cmsswVersion = Tbl(TABLES.CMSSW_VERSION, True)
        
        if self.asc:
            order = "ASC"
        else:
            order = "DESC"


        self.query ="""
        SELECT sub.ds_dataset, sub.rp_proc_version, case sub.rp_hltdebug when 0 then 'Disabled' when 1 then 'Enabled' end rp_hltdebug, case sub.ro_reco when 0 then 'Disabled' when 1 then 'Enabled' end ro_reco, sub.ro_cmssw_version, sub.ro_global_tag, sub.ro_config_url, sub.ro_proc_version
        FROM(
               SELECT
               ds.name ds_dataset,
               rp.proc_version rp_proc_version,
               rp.hltdebug rp_hltdebug,
               ro.do_reco ro_reco,
               sw.name ro_cmssw_version,
               ro.global_tag ro_global_tag,
               ro.config_url ro_config_url,
               ro.proc_version ro_proc_version
               FROM """ + repackConfig + """ rp, """ + recoConfig + """ ro, """ + cmsswVersion  + """ sw, """ + primaryDataset  + """ ds
               WHERE rp.run_id = """ + str(self.runid) + """
               AND rp.run_id = ro.run_id AND rp.dataset_id = ro.dataset_id
               AND ro.cmssw_version_id = sw.id AND rp.dataset_id = ds.dataset_id
               ORDER BY """ + str(int(self.orderby)+1) + " " + order + """
        ) sub
        """
        

class RunsDataProvider(PagedTableDataProvider):
    RUN_ID_TITLE = "Run ID"
    STREAMERS_TITLE = "Streamers"
    REPACKED_TITLE = "Repacked"
    RECO_TITLE = "Reconstructed"
    REPACK_MERGE_TITLE = "Merged"

    cols = Cols([
        Col(RUN_ID_TITLE, decorator=formRunLink),
        Col("Start", decorator=timeToClock),
        Col("End", decorator=timeToClock),
        Col("Process"),
        Col("Version"),
        Col("HLTKey"),
        Col("ACQ_era"),
        Col("Status", info=
            """
                <i style="color:blue;font-size:10px">Active</i> - Run is ongoing and streamer data is still arriving for it<br />
                <i style="color:blue;font-size:10px">CloseOutSegmentInjector</i> - The Tier-0 is trying to make consistent luminosity sections to be repacked (currently we need 4*n streamer files per Luminosity Section)<br />
                <i style="color:blue;font-size:10px">CloseOutScheduler</i> - There is streamer data for the run that has not been assigned to a repack job yet<br />
                <i style="color:blue;font-size:10px">CloseOutRepack</i> - There is streamer data in a run that has not been repacked yet (create repack jobs and wait for them to finish)<br />
                <i style="color:blue;font-size:10px">CloseOutRepackMerge</i> - There is RAW data in the run that needs to be merged (create repack merge jobs and wait from them to finish)<br />
                <i style="color:blue;font-size:10px">CloseOutPromptReco</i> - There is RAW data in the run that needs to be reconstructed (create prompt reco jobs and wait for them to finish)<br />
                <i style="color:blue;font-size:10px">Complete</i> - We have finished processing this run (there could have been failures along the way but we've exhausted all automatic retries)
            """
        ),
        Col(STREAMERS_TITLE, info=
            """
            The number of streamer files
            """
        ),
        Col(REPACKED_TITLE, info=
            """
                The status bar and percentage displays how many streamer
                files were repacked<br /><br />The number shows how many
                repack files were produced
            """
        ),
        Col(REPACK_MERGE_TITLE, info=
            """
                The status bar and percentage displays how many repacked
                files were merged<br /><br />The number shows how many
                merge files were produced
            """
        ),
        Col(RECO_TITLE, info=
            """
                The status bar and percentage displays how many merged
                files, marked for reconstruction, were
                actually reconstructed<br /><br />The number shows how many
                reco files were produced
            """
        ),
    ])

    def __init__(self, orderby = 0, asc = True, page = 0, pageSize = 10):
        PagedTableDataProvider.__init__(self, orderby, asc, page, pageSize)
        self.title = "Runs"
        PagedTableDataProvider._load(self)

    def _analysis(self):
        for line in self.data:
            runid = line[self.cols.getId(self.RUN_ID_TITLE)]
            numStreamers = line[self.cols.getId(self.STREAMERS_TITLE)]
            numRepacked = line[self.cols.getId(self.REPACKED_TITLE)]
            numMerge = line[self.cols.getId(self.REPACK_MERGE_TITLE)]
            numReco = line[self.cols.getId(self.RECO_TITLE)]
            # analyse repacked percentage
            repackedP = formatPercentage(numRepacked, getRunRepackedPercentage(runid, numStreamers))
            # analyse merged percentage
            mergedP = formatPercentage(numMerge, getRunRepackMergePercentage(runid, numRepacked))
            # analyse reco percentage
            recoP = formatPercentage(numReco, getRunRecoPercentage(runid, numReco))

            line[self.cols.getId(self.REPACK_MERGE_TITLE)] = mergedP
            line[self.cols.getId(self.REPACKED_TITLE)] = repackedP
            line[self.cols.getId(self.RECO_TITLE)] = recoP

    def _buildCountQuery(self):
        runTable = Tbl(TABLES.RUN)

        self.countQuery = select(
            [func.count(runTable.c.run_id)]
        )

    def _buildMainQuery(self):
        runTable = Tbl(TABLES.RUN, True)
        runStatusTable = Tbl(TABLES.RUN_STATUS, True)
        streamerTable = Tbl(TABLES.STREAMER, True)
        repackedTable = Tbl(TABLES.REPACKED, True)
        reconstructedTable = Tbl(TABLES.RECONSTRUCTED, True)
        parentageTable = Tbl(TABLES.REPACKED_MERGE_PARENTAGE, True)

        if self.asc: order = "ASC"
        else: order = "DESC"

        # HACK: query time optimisation
        if self.orderby < self.cols.getId(self.STREAMERS_TITLE):
            self.mainQuery = """
            SELECT sub.run_id, sub.start_time, sub.end_time, sub.process, sub.version, sub.hltkey, sub.acq_era, sub.status,
            (SELECT COUNT(*) FROM """ + streamerTable + """ s WHERE s.run_id = sub.run_id) numStreamers,
            (SELECT COUNT(re.run_id) FROM """ + repackedTable + """ re WHERE re.run_id = sub.run_id AND re.repacked_id NOT IN (SELECT output_id FROM """ + parentageTable + """)) numRepacked,
            (SELECT COUNT(*) FROM """ + repackedTable + """ re WHERE (re.status = """ + str(REPACK_RECONSTRUCTION_QUEUED) + """ OR re.status = """ + str(REPACK_RECONSTRUCTION_DISABLED) + """)  AND re.run_id = sub.run_id),
            (SELECT COUNT(*) FROM """ + reconstructedTable + """ rec WHERE rec.run_id = sub.run_id) numReconstructed
            FROM(
                SELECT
                r.run_id run_id,
                r.start_time start_time,
                r.end_time end_time,
                r.process process,
                r.version version,
                r.hltkey hltkey,
                r.acq_era acq_era,
                rs.status status
                FROM """ + runTable + """ r, """ + runStatusTable + """ rs
                WHERE rs.id = r.run_status
                ORDER BY """ + str(int(self.orderby) + 1) + " " + order + """
            ) sub"""
        else:
            # slow query
            self.mainQuery = """
                SELECT
                r.run_id run_id,
                r.start_time start_time,
                r.end_time end_time,
                r.process process,
                r.version version,
                r.hltkey hltkey,
                r.acq_era acq_era,
                rs.status status,
                (SELECT COUNT(s.run_id) FROM """ + streamerTable + """ s WHERE s.run_id = r.run_id) numStreamers,
                (SELECT COUNT(re.run_id) FROM """ + repackedTable + """ re WHERE re.run_id = r.run_id AND re.repacked_id NOT IN (SELECT output_id FROM """ + parentageTable + """)) numRepacked,
                (SELECT COUNT(*) FROM """ + repackedTable + """ re WHERE (re.status = """ + str(REPACK_RECONSTRUCTION_QUEUED) + """ OR re.status = """ + str(REPACK_RECONSTRUCTION_DISABLED) + """)  AND re.run_id = r.run_id),
                (SELECT COUNT(rec.run_id) FROM """ + reconstructedTable + """ rec WHERE rec.run_id = r.run_id) numReconstructed
                FROM """ + runTable + """ r, """ + runStatusTable + """ rs
                WHERE rs.id = r.run_status
                ORDER BY """ + str(int(self.orderby) + 1) + " " + order

#        globals.logger.log(self.query)
